/*  camserv - An internet streaming picture application
 *
 *  Copyright (C) 1999  Jon Travis (jon@dsndata.com)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h> 
#include <stdlib.h>
#include <string.h>
#include <glib.h>

#include "camconfig.h"

struct camconfig_st {
  GHashTable *table;
};

/*
 * camconfig_new:  Create a new camconfig configuration, and initialize
 *                 the default values.  
 *
 * Return Values:  Returns NULL on failure, a valid pointer on success.
 *                 To free the configuration, camconfig_dest should be called.
 */

static 
CamConfig *camconfig_new(){
  CamConfig *res;

  if( (res = malloc( sizeof( *res ))) == NULL )
    return NULL;

  res->table = g_hash_table_new( g_str_hash, g_str_equal );
  if( res->table == NULL ){
    free( res );
    return NULL;
  }

  return res;
}

static
gboolean remove_func( gpointer key, gpointer val, gpointer cldat ){
  free( key );
  free( val );

  return 1;
}

/*
 * camconfig_dest:  Destroy a camconfig structure and all the data
 *                  contained within it (keys and values).
 *
 * Arguments:       cfg = cfg to free
 */

void camconfig_dest( CamConfig *cfg ){
  g_hash_table_foreach_remove( cfg->table, remove_func, NULL );
  g_hash_table_destroy( cfg->table );
  free( cfg );
}

/*
 * camconfig_read:  Read in a camera configuration file into a new
 *                  camconfig structure.  
 * 
 * Arguments:       cfg_file = Configuration file to read
 *
 * Return values:   Returns NULL on failure, else a valid pointer to a cfg
 *                  structure which must be passed to camconfig_dest
 */

CamConfig *camconfig_read( const char *cfg_file ){
  char bigbuf[ 1024 ], key[ 1024 ], value[ 1024 ];
  int line_number;
  CamConfig *res;
  FILE *fp;

  if( (res = camconfig_new()) == NULL )
    return NULL;

  if( (fp = fopen( cfg_file, "r" )) == NULL ){
    camconfig_dest( res );
    return NULL;
  }

  line_number = 0;
  while( fgets( bigbuf, sizeof( bigbuf ), fp ) != NULL ){
    char *keydup = NULL, *valdup = NULL;
    line_number++;

    if( bigbuf[ 0 ] == '#' || bigbuf[ 0 ] == '\n') continue;
    if( sscanf( bigbuf, "%s %[^\n]s", key, value ) != 2 ){
      fprintf( stderr, "camconfig_read: Error on config file line %d\n", 
	       line_number );
      continue;
    }

    if( (keydup = strdup( key )) == NULL ||
	(valdup = strdup( value )) == NULL )
    {
        fprintf( stderr, "camconfig_read: Error allocating memory!\n");
	fclose( fp );
	camconfig_dest( res );
	if( keydup ) free( keydup );
	if( valdup ) free( valdup );
	return NULL;
    }
    g_hash_table_insert( res->table, keydup, valdup );
  }

  return res;
}

/*
 * camconfig_query_int:  Lookup a key/value pair in the configuration
 *                       structure, and return the value as an integer.
 *
 * Arguments:            cfg = Configuration structure containing 'key'
 *                       key = Key to lookup
 *                       err = Location to place whether or not an error
 *                             occurred.  err will be == 1 on failure, and
 *                             0 on success
 *
 * Return values:        Returns the integer representation of the value
 *                       pointed at by 'key'.  atoi() is the translation
 *                       method.
 */

int camconfig_query_int( const CamConfig *cfg, const char *key, int *err ){
  gpointer value;

  value = g_hash_table_lookup( cfg->table, key );
  if( value == NULL ){
    *err = 1;
    return 0;
  }  else {
    *err = 0;
    return atoi( value );
  }
}

/*
 * camconfig_query_def_int:  Lookup a key/value pair in the configuration
 *                           structure, and return the value as an integer.
 *                           If the key/value pair could not be found, then
 *                           a default value is returned.
 *
 * Arguments:            cfg = Configuration structure containing 'key'
 *                       key = Key to lookup
 *                       def = value to return if the lookup fails.
 *
 * Return values:        Returns the integer representation of the value
 *                       pointed at by 'key'.  atoi() is the translation
 *                       method.
 */


int camconfig_query_def_int( const CamConfig *cfg, const char *key, int def ){
  gpointer value;

  value = g_hash_table_lookup( cfg->table, key );
  if( value == NULL ){
    fprintf( stderr, "camconfig:  Using default value of %d for key \"%s\"\n",
	     def, key );
    return def;
  }  else {
    return atoi( value );
  }
}

float camconfig_query_def_float( const CamConfig *cfg, const char *key, 
			       float def )
{
  gpointer value;

  value = g_hash_table_lookup( cfg->table, key );
  if( value == NULL ){
    fprintf( stderr, "camconfig:  Using default value of %f for key \"%s\"\n",
	     def, key );
    return def;
  }  else {
    return atof( value );
  }
}


/*
 * camconfig_query_str:  Lookup a key/value pair in the configuration 
 *                       structure, and return the value as a string.
 *
 * Arguments:            cfg = Configuration structure containing 'key'
 *                       key = Key to lookup
 *
 * Return values:        Returns NULL on failure, else a valid pointer to
 *                       the string value in the config table.  
 */

const char *camconfig_query_str( const CamConfig *cfg, const char *key ){
  gpointer value;

  value = g_hash_table_lookup( cfg->table, key );
  return (const char *)value;
}

/*
 * camconfig_set_str:   Set the value of a string in a config structure.
 *                      If the config structure does not contain the
 *                      key/value pair already, it will be created.
 *
 * Arguments:           cfg = Configuration structure containing 'key'
 *                      key = Key to set the value of
 *                      value = Value to set of 'key'
 *
 * Return values:       Returns -1 if an error occurred (malloc), else 0
 */

int camconfig_set_str( CamConfig *cfg, const char *key, const char *value ){
  gpointer orig_key, orig_value;
  char *new_val = NULL, *new_key = NULL;
  
  if( (new_val = strdup( value )) == NULL )
    return -1;

  if( !g_hash_table_lookup_extended( cfg->table, key, &orig_key, &orig_value))
  {
      /* This is a new key/value datapair */
      if( (new_key = strdup( key )) == NULL ){
	  free( new_val );
	  return -1;
      }

      g_hash_table_insert( cfg->table, new_key, new_val );
      return 0;
  }

  /* Replace an existing value */
  if( orig_value != NULL )
    free( (char *)orig_value );

  g_hash_table_insert( cfg->table, orig_key, new_val  );
  return 0;
}



/*
 * camconfig_set_int:  Set the value of a int in a config structure.
 *                     If the config structure does not contain the 
 *                     key/value pair already, it will be created.
 *
 * Arguments:           cfg = Configuration structure containing 'key'
 *                      key = Key to set the value of
 *                      value = Value to set of 'key'
 *
 * Return values:       Returns -1 if an error occurred (malloc), else 0
 */

int camconfig_set_int( CamConfig *cfg, const char *key, int value ){
  char buf[ 1024 ];
  
  sprintf( buf, "%d", value );
  return camconfig_set_str( cfg, key, buf );
}
